@@ -26,6 +26,8 @@ import (
 	monitoring "github.com/coreos/prometheus-operator/pkg/client/versioned"
 	"github.com/coreos/prometheus-operator/pkg/k8sutil"
 	prometheusoperator "github.com/coreos/prometheus-operator/pkg/prometheus"
+	"github.com/coreos/prometheus-operator/pkg/thanos"
+	thanosoperator "github.com/coreos/prometheus-operator/pkg/thanos"
 	configv1 "github.com/openshift/api/config/v1"
 	routev1 "github.com/openshift/api/route/v1"
 	secv1 "github.com/openshift/api/security/v1"
@@ -35,7 +37,6 @@ import (
 	"github.com/pkg/errors"
 	appsv1 "k8s.io/api/apps/v1"
 	v1 "k8s.io/api/core/v1"
-	"k8s.io/api/extensions/v1beta1"
 	v1betaextensions "k8s.io/api/extensions/v1beta1"
 	rbacv1 "k8s.io/api/rbac/v1"
 	extensionsobj "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
@@ -322,6 +323,26 @@ func (c *Client) CreateOrUpdateAlertmanager(a *monv1.Alertmanager) error {
 	return errors.Wrap(err, "updating Alertmanager object failed")
 }
 
+func (c *Client) CreateOrUpdateThanosRuler(t *monv1.ThanosRuler) error {
+	trclient := c.mclient.MonitoringV1().ThanosRulers(t.GetNamespace())
+	oldTr, err := trclient.Get(t.GetName(), metav1.GetOptions{})
+	if apierrors.IsNotFound(err) {
+		_, err := trclient.Create(t)
+		return errors.Wrap(err, "creating Thanos Ruler object failed")
+	}
+	if err != nil {
+		return errors.Wrap(err, "retrieving Thanos Ruler object failed")
+	}
+
+	t.ResourceVersion = oldTr.ResourceVersion
+	if t.Spec.Storage != nil {
+		t.Spec.Storage.VolumeClaimTemplate.CreationTimestamp = metav1.Unix(0, 0)
+	}
+
+	_, err = trclient.Update(t)
+	return errors.Wrap(err, "updating Thanos Ruler object failed")
+}
+
 func (c *Client) DeleteConfigMap(cm *v1.ConfigMap) error {
 	err := c.kclient.CoreV1().ConfigMaps(cm.GetNamespace()).Delete(cm.GetName(), &metav1.DeleteOptions{})
 	if apierrors.IsNotFound(err) {
@@ -331,38 +352,42 @@ func (c *Client) DeleteConfigMap(cm *v1.ConfigMap) error {
 	return err
 }
 
-func (c *Client) DeleteHashedConfigMap(newHash, prefix string) error {
+// DeleteHashedConfigMap deletes all configmaps in the given namespace which have
+// the specified prefix, and DO NOT have the given hash.
+func (c *Client) DeleteHashedConfigMap(namespace, prefix, newHash string) error {
 	ls := "monitoring.openshift.io/name=" + prefix + ",monitoring.openshift.io/hash!=" + newHash
-	configMaps, err := c.KubernetesInterface().CoreV1().ConfigMaps(c.namespace).List(metav1.ListOptions{
+	configMaps, err := c.KubernetesInterface().CoreV1().ConfigMaps(namespace).List(metav1.ListOptions{
 		LabelSelector: ls,
 	})
 	if err != nil {
-		return errors.Wrapf(err, "error listing configmaps with label selector %s", ls)
+		return errors.Wrapf(err, "error listing configmaps in namespace %s with label selector %s", namespace, ls)
 	}
 
-	for i := range configMaps.Items {
-		err := c.KubernetesInterface().CoreV1().ConfigMaps(c.namespace).Delete(configMaps.Items[i].Name, &metav1.DeleteOptions{})
+	for _, cm := range configMaps.Items {
+		err := c.KubernetesInterface().CoreV1().ConfigMaps(namespace).Delete(cm.Name, &metav1.DeleteOptions{})
 		if err != nil {
-			return errors.Wrapf(err, "error deleting configmap: %s", configMaps.Items[i].Name)
+			return errors.Wrapf(err, "error deleting configmap: %s/%s", namespace, cm.Name)
 		}
 	}
 
 	return nil
 }
 
-func (c *Client) DeleteHashedSecret(newHash, prefix string) error {
+// DeleteHashedSecret deletes all secrets in the given namespace which have
+// the specified prefix, and DO NOT have the given hash.
+func (c *Client) DeleteHashedSecret(namespace, prefix, newHash string) error {
 	ls := "monitoring.openshift.io/name=" + prefix + ",monitoring.openshift.io/hash!=" + newHash
-	configMaps, err := c.KubernetesInterface().CoreV1().Secrets(c.namespace).List(metav1.ListOptions{
+	secrets, err := c.KubernetesInterface().CoreV1().Secrets(namespace).List(metav1.ListOptions{
 		LabelSelector: ls,
 	})
 	if err != nil {
-		return errors.Wrapf(err, "error listing secrets with label selector %s", ls)
+		return errors.Wrapf(err, "error listing secrets in namespace %s with label selector %s", namespace, ls)
 	}
 
-	for i := range configMaps.Items {
-		err := c.KubernetesInterface().CoreV1().Secrets(c.namespace).Delete(configMaps.Items[i].Name, &metav1.DeleteOptions{})
+	for _, s := range secrets.Items {
+		err := c.KubernetesInterface().CoreV1().Secrets(namespace).Delete(s.Name, &metav1.DeleteOptions{})
 		if err != nil {
-			return errors.Wrapf(err, "error deleting secret: %s", configMaps.Items[i].Name)
+			return errors.Wrapf(err, "error deleting secret: %s/%s", namespace, s.Name)
 		}
 	}
 
@@ -409,7 +434,37 @@ func (c *Client) DeletePrometheus(p *monv1.Prometheus) error {
 	return nil
 }
 
-func (c *Client) DeleteDaemonSet(d *v1beta1.DaemonSet) error {
+func (c *Client) DeleteThanosRuler(tr *monv1.ThanosRuler) error {
+	trclient := c.mclient.MonitoringV1().ThanosRulers(tr.GetNamespace())
+
+	err := trclient.Delete(tr.GetName(), nil)
+	if err != nil && !apierrors.IsNotFound(err) {
+		return errors.Wrap(err, "deleting Thanos Ruler object failed")
+	}
+
+	var lastErr error
+	if err := wait.Poll(time.Second*10, time.Minute*10, func() (bool, error) {
+		pods, err := c.KubernetesInterface().CoreV1().Pods(tr.GetNamespace()).List(thanosoperator.ListOptions(tr.GetName()))
+		if err != nil {
+			return false, errors.Wrap(err, "retrieving pods during polling failed")
+		}
+
+		klog.V(6).Infof("waiting for %d Pods to be deleted", len(pods.Items))
+		klog.V(6).Infof("done waiting? %t", len(pods.Items) == 0)
+
+		lastErr = fmt.Errorf("waiting for %d Pods to be deleted", len(pods.Items))
+		return len(pods.Items) == 0, nil
+	}); err != nil {
+		if err == wait.ErrWaitTimeout && lastErr != nil {
+			err = lastErr
+		}
+		return errors.Wrap(err, "waiting for Thanos Ruler Pods to be gone failed")
+	}
+
+	return nil
+}
+
+func (c *Client) DeleteDaemonSet(d *appsv1.DaemonSet) error {
 	orphanDependents := false
 	err := c.kclient.AppsV1().DaemonSets(d.GetNamespace()).Delete(d.GetName(), &metav1.DeleteOptions{OrphanDependents: &orphanDependents})
 	if apierrors.IsNotFound(err) {
@@ -471,6 +526,14 @@ func (c *Client) DeleteService(svc *v1.Service) error {
 	return err
 }
 
+func (c *Client) DeletePrometheusRule(rule *monv1.PrometheusRule) error {
+	err := c.mclient.MonitoringV1().PrometheusRules(rule.GetNamespace()).Delete(rule.GetName(), &metav1.DeleteOptions{})
+	if apierrors.IsNotFound(err) {
+		return nil
+	}
+	return err
+}
+
 func (c *Client) DeleteSecret(s *v1.Secret) error {
 	err := c.kclient.CoreV1().Secrets(s.Namespace).Delete(s.GetName(), &metav1.DeleteOptions{})
 	if apierrors.IsNotFound(err) {
@@ -534,24 +597,65 @@ func (c *Client) WaitForAlertmanager(a *monv1.Alertmanager) error {
 	return nil
 }
 
+func (c *Client) WaitForThanosRuler(t *monv1.ThanosRuler) error {
+	var lastErr error
+	if err := wait.Poll(time.Second*10, time.Minute*5, func() (bool, error) {
+		tr, err := c.mclient.MonitoringV1().ThanosRulers(t.GetNamespace()).Get(t.GetName(), metav1.GetOptions{})
+		if err != nil {
+			return false, errors.Wrap(err, "retrieving Thanos Ruler object failed")
+		}
+		status, _, err := thanos.ThanosRulerStatus(c.kclient.(*kubernetes.Clientset), tr)
+		if err != nil {
+			return false, errors.Wrap(err, "retrieving Thanos Ruler status failed")
+		}
+
+		expectedReplicas := *tr.Spec.Replicas
+		if status.UpdatedReplicas == expectedReplicas && status.AvailableReplicas >= expectedReplicas {
+			return true, nil
+		}
+		lastErr = fmt.Errorf("expected %d replicas, updated %d and available %d", expectedReplicas, status.UpdatedReplicas, status.AvailableReplicas)
+		return false, nil
+	}); err != nil {
+		if err == wait.ErrWaitTimeout && lastErr != nil {
+			err = lastErr
+		}
+		return errors.Wrap(err, "waiting for Thanos Ruler")
+	}
+	return nil
+}
+
 func (c *Client) CreateOrUpdateDeployment(dep *appsv1.Deployment) error {
 	d, err := c.kclient.AppsV1().Deployments(dep.GetNamespace()).Get(dep.GetName(), metav1.GetOptions{})
 
 	if apierrors.IsNotFound(err) {
 		err = c.CreateDeployment(dep)
-		return errors.Wrap(err, "creating deployment object failed")
+		return errors.Wrap(err, "creating Deployment object failed")
 	}
 	if err != nil {
-		return errors.Wrap(err, "retrieving deployment object failed")
+		return errors.Wrap(err, "retrieving Deployment object failed")
 	}
 	if reflect.DeepEqual(dep.Spec, d.Spec) {
 		// Nothing to do, as the currently existing Telemeter client
 		// deployment is equivalent to the one that would be applied.
 		return nil
 	}
-
 	err = c.UpdateDeployment(dep)
-	return errors.Wrap(err, "updating deployment object failed")
+	if err != nil {
+		uErr, ok := err.(*apierrors.StatusError)
+		if ok && uErr.ErrStatus.Code == 422 && uErr.ErrStatus.Reason == metav1.StatusReasonInvalid {
+			// try to delete Deployment
+			err = c.DeleteDeployment(dep)
+			if err != nil {
+				return errors.Wrap(err, "deleting Deployment object failed")
+			}
+			err = c.CreateDeployment(dep)
+			if err != nil {
+				return errors.Wrap(err, "creating Deployment object failed after update failed")
+			}
+		}
+		return errors.Wrap(err, "updating Deployment object failed")
+	}
+	return nil
 }
 
 func (c *Client) CreateDeployment(dep *appsv1.Deployment) error {
@@ -690,7 +794,22 @@ func (c *Client) CreateOrUpdateDaemonSet(ds *appsv1.DaemonSet) error {
 	}
 
 	err = c.UpdateDaemonSet(ds)
-	return errors.Wrap(err, "updating DaemonSet object failed")
+	if err != nil {
+		uErr, ok := err.(*apierrors.StatusError)
+		if ok && uErr.ErrStatus.Code == 422 && uErr.ErrStatus.Reason == metav1.StatusReasonInvalid {
+			// try to delete DaemonSet
+			err = c.DeleteDaemonSet(ds)
+			if err != nil {
+				return errors.Wrap(err, "deleting DaemonSet object failed")
+			}
+			err = c.CreateDaemonSet(ds)
+			if err != nil {
+				return errors.Wrap(err, "creating DaemonSet object failed after update failed")
+			}
+		}
+		return errors.Wrap(err, "updating DaemonSet object failed")
+	}
+	return nil
 }
 
 func (c *Client) CreateDaemonSet(ds *appsv1.DaemonSet) error {
@@ -924,7 +1043,7 @@ func (c *Client) CreateOrUpdateClusterRole(cr *rbacv1.ClusterRole) error {
 
 func (c *Client) CreateOrUpdateClusterRoleBinding(crb *rbacv1.ClusterRoleBinding) error {
 	crbClient := c.kclient.RbacV1().ClusterRoleBindings()
-	_, err := crbClient.Get(crb.GetName(), metav1.GetOptions{})
+	existing, err := crbClient.Get(crb.GetName(), metav1.GetOptions{})
 	if apierrors.IsNotFound(err) {
 		_, err := crbClient.Create(crb)
 		return errors.Wrap(err, "creating ClusterRoleBinding object failed")
@@ -933,7 +1052,21 @@ func (c *Client) CreateOrUpdateClusterRoleBinding(crb *rbacv1.ClusterRoleBinding
 		return errors.Wrap(err, "retrieving ClusterRoleBinding object failed")
 	}
 
-	_, err = crbClient.Update(crb)
+	changed := reflect.DeepEqual(crb.RoleRef, existing.RoleRef)
+	changed = changed || reflect.DeepEqual(crb.Subjects, existing.Subjects)
+	changed = changed || reflect.DeepEqual(crb.Labels, existing.Labels)
+	changed = changed || reflect.DeepEqual(crb.Annotations, existing.Annotations)
+
+	if !changed {
+		return nil
+	}
+
+	err = crbClient.Delete(crb.Name, nil)
+	if err != nil {
+		return errors.Wrap(err, "deleting ClusterRoleBinding object failed")
+	}
+
+	_, err = crbClient.Create(crb)
 	return errors.Wrap(err, "updating ClusterRoleBinding object failed")
 }
 
