@@ -15,7 +15,6 @@
 package operator
 
 import (
-	"fmt"
 	"strings"
 	"time"
 
@@ -36,7 +35,7 @@ import (
 )
 
 const (
-	resyncPeriod = 5 * time.Minute
+	resyncPeriod = 15 * time.Minute
 
 	// see https://github.com/kubernetes/apiserver/blob/b571c70e6e823fd78910c3f5b9be895a756f4cbb/pkg/server/options/authentication.go#L239
 	apiAuthenticationConfigMap    = "kube-system/extension-apiserver-authentication"
@@ -271,7 +270,7 @@ func (o *Operator) processNextWorkItem() bool {
 
 	o.reconcileErrors.Inc()
 	klog.Errorf("Syncing %q failed", key)
-	utilruntime.HandleError(errors.Wrap(err, fmt.Sprintf("sync %q failed", key)))
+	utilruntime.HandleError(errors.Wrapf(err, "sync %q failed", key))
 	o.queue.AddRateLimited(key)
 
 	return true
@@ -326,6 +325,7 @@ func (o *Operator) sync(key string) error {
 			tasks.NewTaskSpec("Updating Telemeter client", tasks.NewTelemeterClientTask(o.client, factory, config)),
 			tasks.NewTaskSpec("Updating configuration sharing", tasks.NewConfigSharingTask(o.client, factory)),
 			tasks.NewTaskSpec("Updating Thanos Querier", tasks.NewThanosQuerierTask(o.client, factory, config.UserWorkloadConfig)),
+			tasks.NewTaskSpec("Updating User Workload Thanos Ruler", tasks.NewThanosRulerUserWorkloadTask(o.client, factory, config.UserWorkloadConfig)),
 		},
 	)
 
